r"""Fake shtab
==============
"""

from argparse import ArgumentParser
from typing import Any

FILE = None
DIRECTORY = DIR = None


def add_argument_to(parser: ArgumentParser, *args: Any, **kwargs: Any):
    from argparse import Action

    Action.complete = None  # type: ignore
    return parser



r"""Provide ``__version__`` for
`importlib.metadata.version() <https://docs.python.org/3/library/importlib.metadata.html#distribution-versions>`_.
"""

from typing import Literal

try:
    from ._version import __version__, __version_tuple__  # type: ignore
except ImportError:  # for setuptools-generate
    __version__ = "rolling"
    __version_tuple__ = (0, 0, 0, __version__, "")

__all__ = ["__version__", "__version_tuple__"]
FILETYPE = Literal["{{ language }}"]



r"""Schema
==========
"""

from collections.abc import Callable
from dataclasses import dataclass
from typing import Any

from lsprotocol.types import Position, Range
from tree_sitter import Node, Tree

from . import UNI


@dataclass
class Trie:
    r"""Trie."""

    range: Range
    parent: "Trie | None" = None
    # can be serialized to a json
    value: dict[str, "Trie"] | list["Trie"] | str | int | float | None = None

    def get_root(self) -> "Trie":
        r"""Get root.

        :rtype: "Trie"
        """
        node = self
        while node.parent is not None:
            node = node.parent
        return node

    def to_path(self) -> str:
        r"""To path.

        :rtype: str
        """
        if self.parent is None:
            return "$"
        path = self.parent.to_path()
        if isinstance(self.parent.value, dict):
            for k, v in self.parent.value.items():
                if v is self:
                    return f"{path}.{k}"
            raise TypeError
        if isinstance(self.parent.value, list):
            for k, v in enumerate(self.parent.value):
                if v is self:
                    return f"{path}[{k}]"
            raise TypeError
        return path

    def from_path(self, path: str) -> "Trie":
        r"""From path.

        :param path:
        :type path: str
        :rtype: "Trie"
        """
        node = self
        if path.startswith("$"):
            path = path.lstrip("$")
            node = self.get_root()
        return node.from_relative_path(path)

    def from_relative_path(self, path: str) -> "Trie":
        r"""From relative path.

        :param path:
        :type path: str
        :rtype: "Trie"
        """
        if path == "":
            return self
        if path.startswith("."):
            if not isinstance(self.value, dict):
                raise TypeError
            path = path.lstrip(".")
            index, mid, path = path.partition(".")
            if mid == ".":
                path = mid + path
            index, mid, suffix = index.partition("[")
            if mid == "[":
                path = mid + suffix + path
            return self.value[index].from_relative_path(path)
        if path.startswith("["):
            if not isinstance(self.value, list):
                raise TypeError
            path = path.lstrip("[")
            index, _, path = path.partition("]")
            return self.value[int(index)].from_relative_path(path)
        raise TypeError

    def to_json(self) -> dict[str, Any] | list[Any] | str | int | float | None:
        r"""To json.

        :rtype: dict[str, Any] | list[Any] | str | int | float | None
        """
        if isinstance(self.value, dict):
            return {k: v.to_json() for k, v in self.value.items()}
        if isinstance(self.value, list):
            return [v.to_json() for v in self.value]
        return self.value

    @classmethod
    def from_tree(cls, tree: Tree) -> "Trie":
        r"""From tree.

        :param tree:
        :type tree: Tree
        :rtype: "Trie"
        """
        return cls.from_node(tree.root_node, None)

    @classmethod
    def from_file(cls, file: str, parse: Callable[[bytes], Tree]) -> "Trie":
        r"""From file.

        :param file:
        :type file: str
        :param parse:
        :type parse: Callable[[bytes], Tree]
        :rtype: "Trie"
        """
        with open(file, "rb") as f:
            text = f.read()
        return cls.from_tree(parse(text))

    @classmethod
    def from_node(cls, node: Node, parent: "Trie | None") -> "Trie":
        r"""From node.

        :param node:
        :type node: Node
        :param parent:
        :type parent: Trie | None
        :rtype: "Trie"
        """
        if parent is None:
            _range = Range(Position(0, 0), Position(1, 0))
        else:
            _range = UNI(node).range
        trie = cls(_range, parent, {})
        trie.value = {
            UNI(child.children[0]).text: cls.from_node(child, trie)
            for child in node.children
        }
        return trie



r"""Diagnose
============

Wrap
``Diagnostic <https://microsoft.github.io/language-server-protocol/specifications/specification-current#diagnostic>``_
to a linter.
"""

import sys
from collections.abc import Callable
from typing import Literal

from lsprotocol.types import Diagnostic, DiagnosticSeverity
from pygls.uris import from_fs_path
from tree_sitter import Tree

from . import Finder
from .utils import get_finders, get_paths


def get_diagnostics_by_finders(
    uri: str, tree: Tree, finders: list[Finder]
) -> list[Diagnostic]:
    r"""Get diagnostics by finders.

    :param uri:
    :type uri: str
    :param tree:
    :type tree: Tree
    :param finders:
    :type finders: list[Finder]
    :rtype: list[Diagnostic]
    """
    return [
        diagnostic
        for finder in finders
        for diagnostic in finder.get_diagnostics(uri, tree)
    ]


def get_diagnostics(
    uri: str,
    tree: Tree,
    classes: list[type[Finder]] | None = None,
    filetype: str | None = None,
) -> list[Diagnostic]:
    r"""Get diagnostics.

    :param uri:
    :type uri: str
    :param tree:
    :type tree: Tree
    :param classes:
    :type classes: list[type[Finder]] | None
    :param filetype:
    :type filetype: str | None
    :rtype: list[Diagnostic]
    """
    finders, finder_classes = get_finders(classes)
    if filetype is None:
        return get_diagnostics_by_finders(uri, tree, finders)
    return [
        diagnostic
        for diagnostic in get_diagnostics_by_finders(
            uri, tree, finders + [cls(filetype) for cls in finder_classes]
        )
    ]


def count_level(
    diagnostics: list[Diagnostic],
    level: DiagnosticSeverity = DiagnosticSeverity.Warning,
) -> int:
    r"""Count level.

    :param diagnostics:
    :type diagnostics: list[Diagnostic]
    :param level:
    :type level: DiagnosticSeverity
    :rtype: int
    """
    return len([
        diagnostic
        for diagnostic in diagnostics
        if diagnostic.severity and diagnostic.severity <= level
    ])


class _Colorama:
    """Colorama."""

    def __getattribute__(self, _: str) -> str:
        """Getattribute.

        :param _:
        :type _: str
        :rtype: str
        """
        return ""


def diagnostics2linter_messages(
    path: str,
    diagnostics: list[Diagnostic],
    color: Literal["auto", "always", "never"] = "auto",
    colors: list[str] | None = None,
) -> list[str]:
    r"""Diagnostics2linter messages.

    :param path:
    :type path: str
    :param diagnostics:
    :type diagnostics: list[Diagnostic]
    :param color:
    :type color: Literal["auto", "always", "never"]
    :param colors:
    :type colors: list[str] | None
    :rtype: list[str]
    """
    from colorama import Fore, init

    init()
    if not sys.stdout.isatty() and color == "auto" or color == "never":
        Fore = _Colorama()
    if colors is None:
        colors = [Fore.RESET, Fore.RED, Fore.YELLOW, Fore.BLUE, Fore.GREEN]
    return [
        (
            f"{Fore.MAGENTA}{path}{Fore.RESET}:"
            f"{Fore.CYAN}{diagnostic.range.start.line + 1}:"
            f"{diagnostic.range.start.character + 1}{Fore.RESET}-"
            f"{Fore.CYAN}{diagnostic.range.end.line + 1}:"
            f"{diagnostic.range.end.character + 1}{Fore.RESET}:"
            f"{colors[diagnostic.severity if diagnostic.severity else 0]}"
            f"{str(diagnostic.severity).split('.')[-1].lower()}{Fore.RESET}"
            f": {diagnostic.message}"
        )
        for diagnostic in diagnostics
    ]


def check_by_finders(
    paths: list[str],
    parse: Callable[[bytes], Tree],
    finders: list[Finder],
    color: Literal["auto", "always", "never"] = "auto",
) -> int:
    r"""Check by finders.

    :param paths:
    :type paths: list[str]
    :param parse:
    :type parse: Callable[[bytes], Tree]
    :param finders:
    :type finders: list[Finder]
    :param color:
    :type color: Literal["auto", "always", "never"]
    :rtype: int
    """
    count = 0
    lines = []
    for path in paths:
        with open(path, "rb") as f:
            src = f.read()
        tree = parse(src)
        uri = from_fs_path(path)
        if uri is None:
            continue
        diagnostics = get_diagnostics_by_finders(uri, tree, finders)
        count += count_level(diagnostics)
        lines += diagnostics2linter_messages(path, diagnostics, color)
    if text := "\n".join(lines):
        print(text)
    return count


def check(
    paths: list[str],
    parse: Callable[[bytes], Tree],
    classes: list[type[Finder]] | None = None,
    get_filetype: Callable[[str], str] | None = None,
    color: Literal["auto", "always", "never"] = "auto",
) -> int:
    r"""Check.

    :param paths:
    :type paths: list[str]
    :param parse:
    :type parse: Callable[[bytes], Tree]
    :param classes:
    :type classes: list[type[Finder]] | None
    :param get_filetype:
    :type get_filetype: Callable[[str], str] | None
    :param color:
    :type color: Literal["auto", "always", "never"]
    :rtype: int
    """
    finders, finder_classes = get_finders(classes)
    if get_filetype is None:
        return check_by_finders(paths, parse, finders, color)
    return sum(
        check_by_finders(
            filepaths,
            parse,
            finders + [cls(filetype) for cls in finder_classes],
            color,
        )
        for filetype, filepaths in get_paths(paths, get_filetype).items()
    )



r"""Test utils."""

import os
from typing import Literal

from lsp_tree_sitter.utils import get_paths

FILETYPE = Literal["python", "c"]


def get_filetype(path: str) -> FILETYPE | Literal[""]:
    r"""Get filetype.

    :param path:
    :type path: str
    :rtype: FILETYPE | Literal[""]
    """
    ext = path.split(os.path.extsep)[-1]
    if ext == "c":
        return "c"
    if ext == "py":
        return "python"
    return ""


class Test:
    r"""Test."""

    @staticmethod
    def test_get_paths() -> None:
        r"""Test get paths.

        :rtype: None
        """
        result = get_paths(["a.c", "b.c", "c.py"], get_filetype)
        expected = {"c": ["a.c", "b.c"], "python": ["c.py"]}
        assert result == expected



r"""Misc
========
"""

from gzip import decompress
from pathlib import Path
from subprocess import check_output  # nosec: B404
from typing import Literal
from urllib import request

from bs4 import BeautifulSoup, FeatureNotFound
from markdown_it import MarkdownIt
from markdown_it.token import Token
from platformdirs import site_data_dir, user_data_dir
from pygls.uris import uri_scheme
from pypandoc import convert_text


def get_data_paths(appname: str) -> list[Path]:
    return [
        Path(d)
        for d in site_data_dir(appname, multipath=True).split(":")
        + [user_data_dir(appname)]
    ]


def get_man(filename: str) -> str:
    r"""Get man.

    :param filename: such as ``make``
    :type filename: str
    :rtype: str
    """
    number = 5
    if filename.find(".") == -1:
        filename += f".{number}"
    else:
        number = int(filename.split(".")[-1])
    filename += "*"
    text = b""
    file = ""
    for path in get_data_paths("man"):
        for file in (path / f"man{number}").glob(filename):
            try:
                with open(file, "rb") as f:
                    text = f.read()
                break
            except Exception:  # nosec: B112
                continue
    if text == b"":
        raise FileNotFoundError
    _, _, ext = str(file).rpartition(".")
    if ext != str(number):
        text = decompress(text)
    return text.decode()


def get_info(filename: str) -> str:
    r"""Get info.

    :param filename: such as ``automake.info-1``
    :type filename: str
    :rtype: str
    """
    filename += "*"
    text = b""
    file = ""
    for path in get_data_paths("info"):
        for file in path.glob(filename):
            try:
                with open(file, "rb") as f:
                    text = f.read()
                break
            except Exception:  # nosec: B112
                continue
    if text == b"":
        raise FileNotFoundError
    _, _, ext = str(file).rpartition(".")
    if not ext.startswith("info"):
        text = decompress(text)
    return text.decode()


def html2soup(html: str) -> BeautifulSoup:
    r"""Html2soup.

    :param html:
    :type html: str
    :rtype: BeautifulSoup
    """
    try:
        soup = BeautifulSoup(html, "lxml")
    except FeatureNotFound:
        soup = BeautifulSoup(html, "html.parser")
    return soup


def get_soup(
    uri: str,
    converter: Literal["pandoc", "groff"] = "pandoc",
    filetype: str = "man",
) -> BeautifulSoup:
    r"""Get soup.

    pandoc doesn't support mdoc.
    `<https://github.com/jgm/pandoc/issues/9056>`_

    :param uri:
    :type uri: str
    :param converter:
    :type converter: Literal["pandoc", "groff"]
    :param filetype:
    :type filetype: str
    :rtype: BeautifulSoup
    """
    if uri_scheme(uri):
        with request.urlopen(uri) as f:  # nosec: B310
            html = f.read()
    else:
        text = get_man(uri)
        if converter == "pandoc":
            html = convert_text(text, "html", filetype)
        else:
            html = check_output(  # nosec: B603 B607
                ["groff", "-m", filetype, "-Thtml"],
                input=text.encode(),
            ).decode()
    return html2soup(html)


def get_md_tokens(filename: str) -> list[Token]:
    r"""Get markdown tokens.

    :param filename:
    :type filename: str
    :rtype: list[Token]
    """
    md = MarkdownIt("commonmark", {})
    text = get_man(filename)
    return md.parse(convert_text(text, "markdown", "man"))



r"""Schema
==========
"""

from dataclasses import dataclass

from lsp_tree_sitter import UNI
from lsp_tree_sitter.schema import Trie
from lsprotocol.types import Position, Range
from tree_sitter import Node

DIRECTIVES = {
    "set_directive",
    "map_directive",
    "unmap_directive",
    "include_directive",
}


@dataclass
class {{ language | title }}Trie(Trie):
    r"""{{ language | title }} Trie."""

    value: dict[str, "Trie"] | list["Trie"] | str | int | float | bool | None

    @classmethod
    def from_node(cls, node: Node, parent: "Trie | None") -> "Trie":
        r"""From node.

        :param node:
        :type node: Node
        :param parent:
        :type parent: Trie | None
        :rtype: "Trie"
        """
        if node.type == "set_directive":
            node = node.children[2]
            _type = node.type
            if _type == "int":
                convert = int
            elif _type == "float":
                convert = float
            elif _type == "bool":

                def convert(x):
                    return x == "true"
            else:

                def convert(x):
                    return x.strip("'\"")

            text = UNI(node).text
            if text != "":
                text = convert(text)
            return cls(UNI(node).range, parent, text)
        if node.type == "map_directive":
            trie = cls(UNI(node).range, parent, {})
            key: Node = node.children[1]
            if key.type == "mode":
                key: Node = key.next_sibling  # type: ignore
            function: Node = key.next_sibling  # type: ignore
            argument = function.next_sibling

            value: dict[str, Trie] = trie.value  # type: ignore
            value["key"] = cls(UNI(key).range, trie, UNI(key).text)  # type: ignore
            value["function"] = cls(
                UNI(function).range, trie, UNI(function).text
            )  # type: ignore
            if argument := function.next_sibling:
                value["argument"] = cls(
                    UNI(argument).range, trie, UNI(argument).text
                )  # type: ignore
            return trie
        if node.type == "unmap_directive":
            key = node.children[1]
            if key.type == "mode":
                key = key.next_sibling  # type: ignore
            return cls(UNI(key).range, parent, UNI(key).text)
        if node.type == "file":
            trie = cls(Range(Position(0, 0), Position(1, 0)), parent, {})
            for directive in DIRECTIVES:
                _type = directive.split("_")[0]
            for child in node.children:
                if child.type not in DIRECTIVES:
                    continue
                _type = child.type.split("_")[0]
                value: dict[str, Trie] = trie.value  # type: ignore
                if _type not in value:
                    trie.value[_type] = cls(  # type: ignore
                        UNI(child).range,
                        trie,
                        {} if _type != "include" else [],
                    )
                subtrie: Trie = trie.value[child.type.split("_")[0]]  # type: ignore
                value: dict[str, Trie] | list[Trie] = subtrie.value  # type: ignore
                if child.type == "set_directive":
                    value: dict[str, Trie]
                    value[UNI(child.children[1]).text] = cls.from_node(
                        child, subtrie
                    )
                elif child.type == "include_directive":
                    value += [  # type: ignore
                        cls(
                            UNI(child.children[1]).range,
                            subtrie,
                            UNI(child.children[1]).text,
                        )
                    ]
                else:
                    mode = "normal"
                    if child.children[1].type == "mode":
                        mode = (
                            UNI(child.children[1]).text
                            .lstrip("[")
                            .rstrip("]")
                        )
                    if mode not in value:
                        value[mode] = cls(
                            Range(Position(0, 0), Position(1, 0)),
                            subtrie,
                            [],
                        )
                    if child.type == "unmap_directive":
                        value[mode].value += [  # type: ignore
                            cls.from_node(child, value[mode])
                        ]
                    else:
                        value[mode].value += [
                            cls.from_node(child, value[mode])
                        ]  # type: ignore
            return trie
        raise NotImplementedError(node.type)



r"""Misc
========
"""

from typing import Any

from .. import FILETYPE


def get_schema(filetype: FILETYPE = "{{ language }}") -> dict[str, Any]:
    r"""Get schema.

    :param filetype:
    :type filetype: FILETYPE
    :rtype: dict[str, Any]
    """
    from .{{ language }} import init_schema

    return init_schema()[filetype]



r"""{{ language | title }}
==========================
"""

from typing import Any

from .._metainfo import SOURCE, project


def init_schema() -> dict[str, Any]:
    r"""Init schema.

    :rtype: dict[str, Any]
    """
    filetype = "{{ language }}"
    schemas = {
        filetype: {
            "$id": (
                f"{SOURCE}/blob/main/"
                f"src/termux_language_server/assets/json/{filetype}.json"
            ),
            "$schema": "http://json-schema.org/draft-07/schema#",
            "$comment": (
                "Don't edit this file directly! It is generated by "
                f"`{project} --generate-schema={filetype}`."
            ),
            "type": "object",
            "additionalProperties": False,
            "properties": {},
        }
    }
    return schemas



r"""Format
==========

Wrap
``Document Formatting <https://microsoft.github.io/language-server-protocol/specifications/specification-current#textDocument_formatting>``_
to a formatter.
"""

from collections.abc import Callable

from lsprotocol.types import Position, Range, TextEdit
from pygls.uris import from_fs_path
from tree_sitter import Tree

from . import Finder
from .utils import get_finders, get_paths


def position_2d_to_1d(source: str, position: Position) -> int:
    r"""Position 2d to 1d.

    :param source:
    :type source: str
    :param position:
    :type position: Position
    :rtype: int
    """
    return (
        sum(len(line) + 1 for line in source.splitlines()[: position.line])
        + position.character
    )


def range_2d_to_1d(source: str, region: Range) -> range:
    r"""Range 2d to 1d.

    :param source:
    :type source: str
    :param region:
    :type region: Range
    :rtype: range
    """
    return range(
        position_2d_to_1d(source, region.start),
        position_2d_to_1d(source, region.end),
    )


def apply_text_edits(text_edits: list[TextEdit], source: str) -> str:
    r"""Apply text edits.

    :param text_edits:
    :type text_edits: list[TextEdit]
    :param source:
    :type source: str
    :rtype: str
    """
    for text_edit in text_edits:
        region = range_2d_to_1d(source, text_edit.range)
        source = (
            source[: region.start] + text_edit.new_text + source[region.stop :]
        )
    return source


def format_by_finders(
    paths: list[str], parse: Callable[[bytes], Tree], finders: list[Finder]
) -> None:
    r"""Format by finders.

    :param paths:
    :type paths: list[str]
    :param parse:
    :type parse: Callable[[bytes], Tree]
    :param finders:
    :type finders: list[Finder]
    :rtype: None
    """
    for path in paths:
        with open(path, "rb") as f:
            src = f.read()
        tree = parse(src)
        uri = from_fs_path(path)
        if uri is None:
            continue
        text_edits = [
            text_edit
            for finder in finders
            for text_edit in finder.get_text_edits(uri, tree)
        ]
        src = apply_text_edits(text_edits, src.decode())
        with open(path, "w") as f:
            f.write(src)


def get_text_edits_by_finders(
    uri: str, tree: Tree, finders: list[Finder]
) -> list[TextEdit]:
    r"""Get text edits by finders.

    :param uri:
    :type uri: str
    :param tree:
    :type tree: Tree
    :param finders:
    :type finders: list[Finder]
    :rtype: list[TextEdit]
    """
    return [
        text_edit
        for finder in finders
        for text_edit in finder.get_text_edits(uri, tree)
    ]


def get_text_edits(
    uri: str,
    tree: Tree,
    classes: list[type[Finder]] | None = None,
    filetype: str | None = None,
) -> list[TextEdit]:
    r"""Get text edits.

    :param uri:
    :type uri: str
    :param tree:
    :type tree: Tree
    :param classes:
    :type classes: list[type[Finder]] | None
    :param filetype:
    :type filetype: str | None
    :rtype: list[TextEdit]
    """
    finders, finder_classes = get_finders(classes)
    if filetype is None:
        return get_text_edits_by_finders(uri, tree, finders)
    return [
        text_edit
        for text_edit in get_text_edits_by_finders(
            uri, tree, finders + [cls(filetype) for cls in finder_classes]
        )
    ]


def format(
    paths: list[str],
    parse: Callable[[bytes], Tree],
    classes: list[type[Finder]] | None = None,
    get_filetype: Callable[[str], str] | None = None,
) -> None:
    r"""Format.

    :param paths:
    :type paths: list[str]
    :param parse:
    :type parse: Callable[[bytes], Tree]
    :param classes:
    :type classes: list[type[Finder]] | None
    :param get_filetype:
    :type get_filetype: Callable[[str], str] | None
    :rtype: None
    """
    finders, finder_classes = get_finders(classes)
    if get_filetype is None:
        return format_by_finders(paths, parse, finders)
    for filetype, filepaths in get_paths(paths, get_filetype).items():
        format_by_finders(
            filepaths,
            parse,
            finders + [cls(filetype) for cls in finder_classes],
        )



r"""LSP Tree sitter
===================
"""

import os
from dataclasses import dataclass, field
from types import TracebackType
from typing import Any, Self

from jinja2 import Template
from lsprotocol.types import (
    Diagnostic,
    DiagnosticSeverity,
    DocumentLink,
    Location,
    Position,
    Range,
    TextEdit,
)
from pygls.uris import to_fs_path
from tree_sitter import Node, Tree, TreeCursor

try:
    from ._version import __version__, __version_tuple__  # type: ignore
except ImportError:  # for setuptools-generate
    __version__ = "rolling"
    __version_tuple__ = (0, 0, 0, __version__, "")


@dataclass
class UNI:
    r"""Unified node identifier.

    ``uri`` and ``path`` is document's uri and path.
    """

    node: Node
    uri: str = ""

    def __str__(self) -> str:
        r"""Str.

        :rtype: str
        """
        return (
            f"{self.text}@{self.uri}:"
            f"{self.node.start_point[0] + 1}:{self.node.start_point[1] + 1}-"
            f"{self.node.end_point[0] + 1}:{self.node.end_point[1]}"
        )

    @property
    def text(self) -> str:
        r"""Text.

        :rtype: str
        """
        return self.node.text.decode() if self.node.text else ""

    @property
    def filepath(self) -> str:
        r"""Filepath.

        :param self:
        :rtype: str
        """
        filepath = to_fs_path(self.uri)
        if filepath is None:
            raise TypeError
        return filepath

    @property
    def path(self) -> str:
        r"""Path.

        :rtype: str
        """
        text = os.path.expandvars(os.path.expanduser(self.text))
        return os.path.join(os.path.dirname(self.filepath), text)

    @property
    def range(self) -> Range:
        r"""Range.

        :rtype: Range
        """
        return Range(
            Position(*self.node.start_point), Position(*self.node.end_point)
        )

    @property
    def location(self) -> Location:
        r"""Location.

        :rtype: Location
        """
        return Location(self.uri, self.range)

    def get_diagnostic(
        self,
        message: str,
        severity: DiagnosticSeverity,
        **kwargs: Any,
    ) -> Diagnostic:
        r"""Get diagnostic.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :param kwargs:
        :type kwargs: Any
        :rtype: Diagnostic
        """
        return Diagnostic(
            self.range,
            Template(message).render(uni=self, **kwargs),
            severity,
        )

    def get_text_edit(self, new_text: str) -> TextEdit:
        r"""Get text edit.

        :param new_text:
        :type new_text: str
        :rtype: TextEdit
        """
        return TextEdit(self.range, new_text)

    def get_document_link(
        self, target: str = "{{uni.uri}}", **kwargs: Any
    ) -> DocumentLink:
        r"""Get document link.

        :param target:
        :type target: str
        :param kwargs:
        :type kwargs: Any
        :rtype: DocumentLink
        """
        return DocumentLink(
            self.range,
            Template(target).render(uni=self, **kwargs),
        )


@dataclass
class Finder:
    r"""Finder."""

    message: str = ""
    severity: DiagnosticSeverity = DiagnosticSeverity.Error
    max_level: int = 5
    level: int = 0
    unis: list[UNI] = field(default_factory=list)

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        return True

    def __and__(self, second: Self) -> "AndFinder":
        r"""And.

        :param second:
        :type second: Self
        :rtype: AndFinder
        """
        return AndFinder(finders=(self, second))

    def __or__(self, second: Self) -> "OrFinder":
        r"""Or.

        :param second:
        :type second: Self
        :rtype: OrFinder
        """
        return OrFinder(finders=(self, second))

    def __minus__(self, second: Self) -> "AndFinder":
        r"""Minus.

        :param second:
        :type second: Self
        :rtype: AndFinder
        """
        return AndFinder(finders=(self, NotFinder(finder=second)))

    def __invert__(self) -> "NotFinder":
        r"""Invert.

        :rtype: NotFinder
        """
        return NotFinder(finder=self)

    def __enter__(self) -> None:
        self.level += 1

    def __exit__(
        self,
        type: type[BaseException] | None,
        value: BaseException | None,
        tb: TracebackType | None,
    ) -> None:
        self.level -= 1

    def uni2diagnostic(self, uni: UNI) -> Diagnostic:
        r"""Uni2diagnostic.

        :param uni:
        :type uni: UNI
        :rtype: Diagnostic
        """
        return uni.get_diagnostic(self.message, self.severity)

    def unis2diagnostics(self, unis: list[UNI]) -> list[Diagnostic]:
        r"""Unis2diagnostics.

        :param unis:
        :type unis: list[UNI]
        :rtype: list[Diagnostic]
        """
        return [self.uni2diagnostic(uni) for uni in unis]

    def is_include_node(self, node: Node) -> bool:
        r"""Is include node.

        :param node:
        :type node: Node
        :rtype: bool
        """
        return False

    def parse(self, code: bytes) -> Tree:
        r"""Parse.

        :param code:
        :type code: bytes
        :rtype: Tree
        """
        raise NotImplementedError

    def uri2tree(self, uri: str) -> Tree | None:
        r"""Convert ``uri`` to ``Tree``. Used by ``prepare()``,
        ``move_cursor()``.

        :param uri:
        :type uri: str
        :rtype: Tree | None
        """
        path = to_fs_path(uri)
        if path is None:
            raise TypeError
        if not os.path.exists(path):
            return None
        with open(path, "rb") as f:
            code = f.read()
        return self.parse(code)

    def uni2uri(self, uni: UNI) -> str:
        r"""Convert UNI to node's URI. Used by ``move_cursor()`` to parse
        recursively. Override it if necessary. *Visitor pattern*.

        :param uni:
        :type uni: UNI
        :rtype: str
        """
        return uni.uri

    def move_cursor(
        self, uri: str, cursor: TreeCursor, is_all: bool = False
    ) -> str | None:
        r"""Move cursor to next ``UNI``'s ``node`` which can make
        ``self.__call__()`` return ``True``. Then return ``UNI``'s ``uri``. If
        ``self.is_include_node()`` always be ``False``, returned inputted
        ``uri``. If no any ``UNI``'s ``node`` which can make ``self.__call__()
        `` return ``True``, return ``None``.

        :param self:
        :param uri:
        :type uri: str
        :param cursor:
        :type cursor: TreeCursor
        :param is_all:
        :type is_all: bool
        :rtype: str | None
        """
        if cursor.node is None:
            raise TypeError
        while self(UNI(cursor.node, uri)) is False:
            if (
                self.is_include_node(cursor.node)
                and self.level < self.max_level
            ):
                uni = UNI(cursor.node, uri)
                new_uri = self.uni2uri(uni)
                tree = self.uri2tree(new_uri)
                # skip if cannot convert `uni`'s `include_node` to a tree
                if tree is not None:
                    with self:
                        if is_all:
                            # don't `self.reset()` to use `self.unis` again
                            self.find_all(new_uri, tree, False)
                        else:
                            result = self.find(new_uri, tree)
                            if result is not None:
                                return new_uri
            if cursor.node.child_count > 0:
                cursor.goto_first_child()
                continue
            while cursor.node.next_sibling is None:
                cursor.goto_parent()
                # when cannot find new nodes, return
                if cursor.node.parent is None:
                    return None
            cursor.goto_next_sibling()
        return uri

    def reset(self) -> None:
        r"""Reset.

        :rtype: None
        """
        self.level = 0
        self.unis = []

    def prepare(
        self, uri: str, tree: Tree | None = None, reset: bool = True
    ) -> Tree:
        r"""Prepare.

        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree | None
        :param reset:
        :type reset: bool
        :rtype: Tree
        """
        if reset:
            self.reset()
        if tree is None:
            tree = self.uri2tree(uri)
        if tree is None:
            raise TypeError
        return tree

    def find(
        self, uri: str, tree: Tree | None = None, reset: bool = True
    ) -> UNI | None:
        r"""Find.

        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree | None
        :param reset:
        :type reset: bool
        :rtype: UNI | None
        """
        cursor = self.prepare(uri, tree, reset).walk()
        if cursor.node is None:
            raise TypeError
        _uri = self.move_cursor(uri, cursor, False)
        if _uri is not None:
            return UNI(cursor.node, _uri)

    def find_all(
        self, uri: str, tree: Tree | None = None, reset: bool = True
    ) -> list[UNI]:
        r"""Find all.

        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree | None
        :param reset:
        :type reset: bool
        :rtype: list[UNI]
        """
        cursor = self.prepare(uri, tree, reset).walk()
        if cursor.node is None:
            raise TypeError
        while True:
            _uri = self.move_cursor(uri, cursor, True)
            if _uri is not None:
                self.unis += [UNI(cursor.node, _uri)]
            while cursor.node.next_sibling is None:
                cursor.goto_parent()
                # when cannot find new nodes, return
                if cursor.node.parent is None:
                    return self.unis
            cursor.goto_next_sibling()

    def get_diagnostics(self, uri: str, tree: Tree) -> list[Diagnostic]:
        r"""Get diagnostics.

        :param self:
        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree
        :rtype: list[Diagnostic]
        """
        return self.unis2diagnostics(self.find_all(uri, tree))

    def get_text_edits(self, uri: str, tree: Tree) -> list[TextEdit]:
        r"""Get text edits.

        :param self:
        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree
        :rtype: list[TextEdit]
        """
        self.find_all(uri, tree)
        return []

    def get_document_links(
        self, uri: str, tree: Tree, template: str = "{{uni.uri}}"
    ) -> list[DocumentLink]:
        r"""Get document links.

        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree
        :param template:
        :type template: str
        :rtype: list[DocumentLink]
        """
        self.find_all(uri, tree)
        return [
            uni.get_document_link(template) for uni in self.find_all(uri, tree)
        ]


@dataclass
class AndFinder(Finder):
    finders: tuple[Finder, ...] = ()

    def __call__(self, uni: UNI) -> bool:
        return all(finder(uni) for finder in self.finders)


@dataclass
class OrFinder(Finder):
    finders: tuple[Finder, ...] = ()

    def __call__(self, uni: UNI) -> bool:
        return any(finder(uni) for finder in self.finders)


@dataclass
class NotFinder(Finder):
    finder: Finder = field(default_factory=Finder)

    def __call__(self, uni: UNI) -> bool:
        return not self.finder(uni)



r"""Test utils."""

import os

from {{ module }}.finders import Import{{ language | title }}Finder
from {{ module }}.utils import get_schema, parser


class Test:
    r"""Test."""

    @staticmethod
    def test_get_schema() -> None:
        r"""Test get schema.

        :rtype: None
        """
        assert len(
            get_schema()
            .get("properties", {})
            .get("set", {})
            .get("description", "")
            .splitlines()
        )



r"""Utils
=========

Some common functions used by formatters and linters.
"""

import os
import sys
from collections.abc import Callable
from typing import Any, Literal

from . import Finder


def get_paths(
    paths: list[str], get_filetype: Callable[[str], str]
) -> dict[str, list[str]]:
    r"""Get paths.

    :param paths:
    :type paths: list[str]
    :param get_filetype: return ``Literal["ft1", "ft2", ...] | Literal[""]``
    :type get_filetype: Callable[[str], str]
    :rtype: dict[str, list[str]]
    """
    filetype_paths = {
        k: []
        for k in get_filetype.__annotations__["return"].__args__[0].__args__
        if k != ""
    }
    for path in paths:
        filetype = get_filetype(path)
        for _filetype, filepaths in filetype_paths.items():
            if filetype == _filetype:
                filepaths += [path]
    return filetype_paths


def get_finders(
    classes: list[type[Finder]] | None = None,
) -> tuple[list[Finder], list[type[Finder]]]:
    r"""Get finders.

    :param classes:
    :type classes: list[type[Finder]] | None
    :rtype: tuple[list[Finder], list[type[Finder]]]
    """
    if classes is None:
        from .finders import ErrorFinder

        classes = [ErrorFinder]

    finders = []
    finder_classes = []
    for cls in classes:
        if cls.__init__.__annotations__.get("filetype"):
            finder_classes += [cls]
        else:
            finders += [cls()]
    return finders, finder_classes


def pygmentize(text: str, filetype: str) -> None:
    r"""Pygmentize.

    :param text:
    :type text: str
    :param filetype:
    :type filetype: str
    :rtype: None
    """
    TERM = os.getenv("TERM", "xterm")
    if not sys.stdout.isatty():
        TERM = "dumb"
    try:
        from pygments import highlight
        from pygments.formatters import get_formatter_by_name
        from pygments.lexers import get_lexer_by_name

        if TERM.split("-")[-1] == "256color":
            formatter_name = "terminal256"
        elif TERM != "dumb":
            formatter_name = "terminal"
        else:
            formatter_name = None
        if formatter_name:
            formatter = get_formatter_by_name(formatter_name)
            lexer = get_lexer_by_name(filetype)
            print(highlight(text, lexer, formatter), end="")
    except ImportError:
        TERM = "dumb"
    if TERM == "dumb":
        print(text)


def pprint(
    obj: object,
    filetype: Literal["json", "yaml", "toml"] = "json",
    *args: Any,
    **kwargs: Any,
) -> None:
    r"""Pprint.

    :param obj:
    :type obj: object
    :param filetype:
    :type filetype: Literal["json", "yaml", "toml"]
    :param args:
    :type args: Any
    :param kwargs:
    :type kwargs: Any
    :rtype: None
    """
    if filetype == "json":
        from json import dumps
    elif filetype == "yaml":
        from yaml import dump as dumps
    elif filetype == "toml":
        from tomli_w import dumps

    text = dumps(obj, *args, **kwargs)
    pygmentize(text, filetype)



r"""Finders
===========
"""

from dataclasses import dataclass

from lsp_tree_sitter.finders import ErrorFinder, QueryFinder, SchemaFinder
from lsprotocol.types import DiagnosticSeverity

from .schema import {{ language | title }}Trie
from .utils import get_query, get_schema


@dataclass(init=False)
class Import{{ language | title }}Finder(QueryFinder):
    r"""Import{{ language | title }}Finder."""

    def __init__(
        self,
        message: str = "\{\{uni.text\}\}: error",
        severity: DiagnosticSeverity = DiagnosticSeverity.Information,
    ):
        r"""Init.

        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        """
        super().__init__(get_query("import"), message, severity)


@dataclass(init=False)
class {{ language | title }}Finder(SchemaFinder):
    r"""{{ language | title }}finder."""

    def __init__(self) -> None:
        r"""Init.

        :rtype: None
        """
        super().__init__(get_schema(), {{ language | title }}Trie)


DIAGNOSTICS_FINDER_CLASSES = [
    ErrorFinder,
    {{ language | title }}Finder,
]
FORMAT_FINDER_CLASSES = []



r"""Complete
============
"""

import os
from glob import glob
from pathlib import Path
from typing import Any

from lsprotocol.types import (
    CompletionItem,
    CompletionItemKind,
    CompletionList,
    MarkupContent,
    MarkupKind,
)
from pygls.uris import to_fs_path


def get_completion_list_by_enum(
    text: str, property: dict[str, Any]
) -> CompletionList:
    r"""Get completion list by enum. ``property.items`` must contains
    ``enum``, ``oneOf.enum``, ``anyOf.enum`` or ``allOf``.

    :param text:
    :type text: str
    :param property:
    :type property: dict[str, Any]
    :rtype: CompletionList
    """
    # if contains .items, it is an array
    property = property.get("items", property)
    enum = property.get(
        "enum",
        property.get(
            "oneOf", property.get("anyOf", property.get("allOf", [{}]))
        )[0].get("enum", []),
    )
    items = []
    for k in enum:
        if k is None:
            continue
        if not isinstance(k, str):
            k = str(k)
        if k.startswith(text):
            items += [
                CompletionItem(
                    k,
                    kind=CompletionItemKind.Constant,
                    insert_text=k,
                )
            ]
    return CompletionList(False, items)


def get_completion_list_by_uri(
    text: str, uri: str, exprs: dict[str, str] | None = None
) -> CompletionList:
    r"""Get completion list by ``uri``. Don't need to filter by ``text``
    because all results are started with ``text``.

    :param text:
    :type text: str
    :param uri:
    :type uri: str
    :param exprs:
    :type exprs: dict[str, str] | None
    :rtype: CompletionList
    """
    if exprs is None:
        exprs = {"*": "text", "**/*": "text"}
    p = to_fs_path(uri)
    if p is None:
        raise TypeError
    dirname = os.path.dirname(p)
    prefix = os.path.join(dirname, text)
    items = []
    for expr, filetype in exprs.items():
        for file in glob(os.path.join(dirname, expr), recursive=True):
            if not file.startswith(prefix):
                continue
            if os.path.isdir(file):
                file += os.path.sep
            items += [
                CompletionItem(
                    file.rpartition(dirname + os.path.sep)[-1],
                    kind=CompletionItemKind.File
                    if os.path.isfile(file)
                    else CompletionItemKind.Folder,
                    documentation=MarkupContent(
                        MarkupKind.Markdown,
                        f"""```{filetype}
{Path(file).read_text()}
```""",
                    )
                    if os.path.isfile(file)
                    else "\n".join(os.listdir(file)),
                    insert_text=file.rpartition(dirname + os.path.sep)[-1],
                )
            ]
    return CompletionList(False, items)



"""This file is generated by setuptools-generate.
The information comes from pyproject.toml.
It provide some metainfo for docs/conf.py to build documents and
help2man to build man pages.
"""

# For docs/conf.py
project = "lsp-tree-sitter"
author = """Wu, Zhenyu <wuzhenyu@ustc.edu> """
copyright = "2025"

# For help2man
DESCRIPTION = "a library to create language servers"
EPILOG = "Report bugs to https://github.com/neomutt/lsp-tree-sitter/issues"
# format __version__ by yourself
VERSION = """lsp-tree-sitter {__version__}
Copyright (C) 2025
Written by Wu, Zhenyu <wuzhenyu@ustc.edu> """


r"""Finders
===========
"""

import os
from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any

from jinja2 import Template
from jsonschema.protocols import Validator
from jsonschema.validators import validator_for
from lsprotocol.types import (
    Diagnostic,
    DiagnosticSeverity,
    Location,
    Position,
    Range,
    TextEdit,
)
from tree_sitter import Node, Query, QueryCursor, Tree

from . import UNI, Finder
from .schema import Trie


@dataclass
class ErrorFinder(Finder):
    r"""Errorfinder."""

    message: str = "{{uni.text}}: error"
    severity: DiagnosticSeverity = DiagnosticSeverity.Error

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        node = uni.node
        return node.has_error and not (
            any(child.has_error for child in node.children)
        )


@dataclass
class MissingFinder(Finder):
    r"""Missingfinder."""

    message: str = "{{uni.text}}: missing"
    severity: DiagnosticSeverity = DiagnosticSeverity.Error

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        node = uni.node
        return node.is_missing and not (
            any(child.is_missing for child in node.children)
        )


@dataclass
class NotFileFinder(Finder):
    r"""NotFilefinder."""

    message: str = "{{uni.text}}: no such file or directory"
    severity: DiagnosticSeverity = DiagnosticSeverity.Error

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        path = uni.path
        return not (os.path.isfile(path) or os.path.isdir(path))


@dataclass
class RepeatedFinder(Finder):
    r"""Repeatedfinder."""

    message: str = "{{uni.text}}: is repeated on {{_uni}}"
    severity: DiagnosticSeverity = DiagnosticSeverity.Warning
    repeated_unis: list[UNI] = field(default_factory=list)
    uni_pairs: list[tuple[UNI, UNI]] = field(default_factory=list)

    def reset(self) -> None:
        r"""Reset.

        :rtype: None
        """
        self.level = 0
        self.unis = []
        self.repeated_unis = []
        self.uni_pairs = []

    def filter(self, uni: UNI) -> bool:
        r"""Filter.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        return True

    def compare(self, uni: UNI, _uni: UNI) -> bool:
        r"""Compare.

        :param uni:
        :type uni: UNI
        :param _uni:
        :type _uni: UNI
        :rtype: bool
        """
        return uni.node.text == _uni.node.text

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        if self.filter(uni) is False:
            return False
        for _uni in self.repeated_unis:
            if self.compare(uni, _uni):
                self.uni_pairs += [(uni, _uni)]
                return True
        self.repeated_unis += [uni]
        return False

    def get_definitions(self, uni: UNI) -> list[Location]:
        r"""Get definitions.

        :param uni:
        :type uni: UNI
        :rtype: list[Location]
        """
        for uni_, _uni in self.uni_pairs:
            # cache hit
            if uni == uni_:
                return [_uni.location]
        return []

    def get_references(self, uni: UNI) -> list[Location]:
        r"""Get references.

        :param uni:
        :type uni: UNI
        :rtype: list[Location]
        """
        locations = []
        for uni_, _uni in self.uni_pairs:
            # cache hit
            if uni == _uni:
                locations += [uni_.location]
        return locations

    def get_text_edits(self, uri: str, tree: Tree) -> list[TextEdit]:
        r"""Get text edits. Only return two to avoid `Overlapping edit`

        :param self:
        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree
        :rtype: list[TextEdit]
        """
        self.find_all(uri, tree)
        for uni, _uni in self.uni_pairs:
            # swap 2 unis
            return [
                uni.get_text_edit(_uni.text),
                _uni.get_text_edit(uni.text),
            ]
        return []

    def uni2diagnostic(self, uni: UNI) -> Diagnostic:
        r"""Uni2diagnostic.

        :param uni:
        :type uni: UNI
        :rtype: Diagnostic
        """
        for uni_, _uni in self.uni_pairs:
            if uni == uni_:
                return uni.get_diagnostic(
                    self.message, self.severity, _uni=_uni
                )
        return uni.get_diagnostic(self.message, self.severity)


@dataclass
class UnsortedFinder(RepeatedFinder):
    r"""Unsortedfinder."""

    message: str = "{{uni.text}}: is unsorted due to {{_uni}}"
    severity: DiagnosticSeverity = DiagnosticSeverity.Warning

    def compare(self, uni: UNI, _uni: UNI) -> bool:
        r"""Compare.

        :param uni:
        :type uni: UNI
        :param _uni:
        :type _uni: UNI
        :rtype: bool
        """
        if uni.node.text and _uni.node.text:
            return uni.node.text < _uni.node.text
        return True


@dataclass(init=False)
class UnFixedOrderFinder(RepeatedFinder):
    r"""Unfixedorderfinder."""

    def __init__(
        self,
        order: list[Any],
        message: str = "{{uni.text}}: is unsorted due to {{_uni}}",
        severity: DiagnosticSeverity = DiagnosticSeverity.Warning,
    ) -> None:
        r"""Init.

        :param order:
        :type order: list[Any]
        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)
        self.order = order

    def filter(self, uni: UNI) -> bool:
        r"""Filter.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        return uni.text in self.order

    def compare(self, uni: UNI, _uni: UNI) -> bool:
        r"""Compare.

        :param uni:
        :type uni: UNI
        :param _uni:
        :type _uni: UNI
        :rtype: bool
        """
        return self.order.index(uni.text) < self.order.index(_uni.text)


@dataclass(init=False)
class TypeFinder(Finder):
    r"""Typefinder."""

    def __init__(
        self,
        type: str,
        message: str = "",
        severity: DiagnosticSeverity = DiagnosticSeverity.Information,
    ) -> None:
        r"""Init.

        :param type:
        :type type: str
        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)
        self.type = type

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        node = uni.node
        return node.type == self.type


@dataclass(init=False)
class PositionFinder(Finder):
    r"""Positionfinder."""

    def __init__(
        self,
        position: Position,
        left_equal: bool = True,
        right_equal: bool = False,
        message: str = "",
        severity: DiagnosticSeverity = DiagnosticSeverity.Information,
    ) -> None:
        r"""Init.

        :param position:
        :type position: Position
        :param left_equal:
        :type left_equal: bool
        :param right_equal:
        :type right_equal: bool
        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)
        self.position = position
        self.left_equal = left_equal
        self.right_equal = right_equal

    @staticmethod
    def belong(
        position: Position,
        node: Node,
        left_equal: bool = True,
        right_equal: bool = False,
    ) -> bool:
        r"""Belong.

        :param position:
        :type position: Position
        :param node:
        :type node: Node
        :param left_equal:
        :type left_equal: bool
        :param right_equal:
        :type right_equal: bool
        :rtype: bool
        """
        if left_equal:
            left_flag = Position(*node.start_point) <= position
        else:
            left_flag = Position(*node.start_point) < position
        if right_equal:
            right_flag = position <= Position(*node.end_point)
        else:
            right_flag = position < Position(*node.end_point)
        return left_flag and right_flag

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        node = uni.node
        return node.child_count == 0 and self.belong(
            self.position, node, self.left_equal, self.right_equal
        )


@dataclass(init=False)
class RangeFinder(Finder):
    r"""Rangefinder."""

    def __init__(
        self,
        range: Range,
        message: str = "",
        severity: DiagnosticSeverity = DiagnosticSeverity.Information,
    ) -> None:
        r"""Init.

        :param range:
        :type range: Range
        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        super().__init__(message, severity)
        self.range = range

    @staticmethod
    def equal(_range: Range, node: Node) -> bool:
        r"""Equal.

        :param _range:
        :type _range: Range
        :param node:
        :type node: Node
        :rtype: bool
        """
        return _range.start == Position(
            *node.start_point
        ) and _range.end == Position(*node.end_point)

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        node = uni.node
        return self.equal(self.range, node)


@dataclass(init=False)
class RequiresFinder(Finder):
    r"""Requiresfinder."""

    def __init__(
        self,
        requires: set[Any],
        message: str = "{{require}}: required",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param requires:
        :type requires: set[Any]
        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        self.initial_requires = requires
        self.requires = deepcopy(self.initial_requires)
        super().__init__(message, severity)

    def reset(self) -> None:
        r"""Reset.

        :rtype: None
        """
        self.level = 0
        self.unis = []
        self.requires = deepcopy(self.initial_requires)

    def filter(self, uni: UNI, require: Any) -> bool:
        r"""Filter.

        :param uni:
        :type uni: UNI
        :param require:
        :type require: Any
        :rtype: bool
        """
        return False

    def __call__(self, uni: UNI) -> bool:
        r"""Call.

        :param uni:
        :type uni: UNI
        :rtype: bool
        """
        found = set()
        for require in self.requires:
            if self.filter(uni, require):
                found |= {require}
        self.requires -= found
        return False

    def require2message(self, require: Any, **kwargs: Any) -> str:
        r"""Require2message.

        :param require:
        :type require: Any
        :param kwargs:
        :type kwargs: Any
        :rtype: str
        """
        return Template(self.message).render(
            uni=self, require=require, **kwargs
        )

    def get_diagnostics(self, uri: str, tree: Tree) -> list[Diagnostic]:
        r"""Get diagnostics.

        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree
        :rtype: list[Diagnostic]
        """
        self.find_all(uri, tree)
        return [
            Diagnostic(
                # If you want to specify a range that contains a line including
                # the line ending character(s) then use an end position
                # denoting the start of the next line
                Range(Position(0, 0), Position(1, 0)),
                self.require2message(i),
                self.severity,
            )
            for i in self.requires
        ]


@dataclass(init=False)
class SchemaFinder(Finder):
    r"""Schemafinder."""

    def __init__(self, schema: dict[str, Any], cls: type[Trie]) -> None:
        r"""Init.

        :param schema:
        :type schema: dict[str, Any]
        :param cls:
        :type cls: type[Trie]
        :rtype: None
        """
        self.validator = self.schema2validator(schema)
        self.cls = cls

    @staticmethod
    def schema2validator(schema: dict[str, Any]) -> Validator:
        r"""Schema2validator.

        :param schema:
        :type schema: dict[str, Any]
        :rtype: Validator
        """
        return validator_for(schema)(schema)

    def get_diagnostics(self, uri: str, tree: Tree) -> list[Diagnostic]:
        r"""Get diagnostics.

        :param _:
        :type _: str
        :param tree:
        :type tree: Tree
        :rtype: list[Diagnostic]
        """
        trie = self.cls.from_tree(tree)
        return [
            Diagnostic(
                trie.from_path(error.json_path).range,
                error.message,
                DiagnosticSeverity.Error,
            )
            for error in self.validator.iter_errors(trie.to_json())
        ]


@dataclass(init=False)
class QueryFinder(Finder):
    r"""Queryfinder."""

    def __init__(
        self,
        query: Query,
        message: str = "",
        severity: DiagnosticSeverity = DiagnosticSeverity.Error,
    ) -> None:
        r"""Init.

        :param query:
        :type query: Query
        :param message:
        :type message: str
        :param severity:
        :type severity: DiagnosticSeverity
        :rtype: None
        """
        self.cursor = QueryCursor(query)
        super().__init__(message, severity)

    def find_all(
        self, uri: str, tree: Tree | None = None, reset: bool = True
    ) -> list[UNI]:
        r"""Find all.

        :param uri:
        :type uri: str
        :param tree:
        :type tree: Tree | None
        :param reset:
        :type reset: bool
        :rtype: list[UNI]
        """
        tree = self.prepare(uri, tree, reset)
        captures = self.cursor.captures(tree.root_node)
        return self.captures2unis(captures, uri)

    def captures2unis(
        self, captures: dict[str, list[Node]], uri: str
    ) -> list[UNI]:
        r"""Captures2unis.

        :param captures:
        :type captures: dict[str, list[Node]]
        :param uri:
        :type uri: str
        :rtype: list[UNI]
        """
        unis = []
        for label, nodes in captures.items():
            if uni := self.capture2uni(label, nodes, uri):
                unis += [uni]
        return unis

    def capture2uni(
        self, label: str, nodes: list[Node], uri: str
    ) -> UNI | None:
        r"""Capture2uni. can return ``None`` to skip.

        :param label:
        :type label: str
        :param nodes:
        :type nodes: list[Node]
        :param uri:
        :type uri: str
        :rtype: UNI | None
        """
        return UNI(nodes[0], uri)
